#!/bin/sh

is_ok=0
is_fail=0
keep=0
verbose=0

if_verbose_2(){
  if test "$verbose" -ge 2; then
    echo $@ 1>&2
  fi
}

pass_verbose(){
  if test "$verbose" -gt 0; then
    echo "--verbose $verbose"
  fi
}

# test suites functions
apply_test(){
  local test_name=$1
  local test_func=$2
  local ref_func=$3
  local cleanup=$4

  $cleanup
  rm -f "$test_name.out" "$test_name.ref"

  if_verbose_2 -n "$test_name: "

  $test_func $test_name > "$test_name.out"
  $ref_func $test_name > "$test_name.ref"
  if diff "$test_name.out" "$test_name.ref" > /dev/null 2>&1; then
    is_ok=$(($is_ok + 1))
    if_verbose_2 "Ok"
  else
    is_fail=$(($is_fail + 1))
    if_verbose_2 "Ko"
  fi

  if test "$keep" -lt 1; then
    $cleanup
    rm -f "$test_name.out" "$test_name.ref"
  fi
}

report_status(){
  if_verbose_2 "successes: $is_ok"
  if_verbose_2 "failures: $is_fail"
  if test "$is_fail" -gt 0; then
    exit 1;
  fi
}

## specialized test suite.
# cleaunp
nothing(){
  echo "nothing" > /dev/null;
}

# use xtc to find a tool
xtc_tool(){
  local tool_name=$1
  xtc query -t $tool_name |\
      sed -e 's/^.*: //'
}

# remove attributes from an aterm tree.
remove_attributes(){
  sed -e 's/{[^}]*}//g'
}

# the testing program.
test_func(){
  local name=$1
  $(xtc_tool "java-fold")				\
      -i "$SRCDIR/$name.java"				\
      -f "$SRCDIR/$name.fold.pl" $(pass_verbose) |	\
      remove_attributes
}

# the reference output.
ref_func(){
  local name=$1
  $(xtc_tool "dryad-front") --tc on	\
      -i "$SRCDIR/$name.res.java" |	\
      remove_attributes
}

# the function used to apply the test.
test_fold(){
  local name=$1
  apply_test "$name" test_func ref_func nothing
}

####
#### Test Suite
####
name=$0

main(){
  is_ok=0
  is_fail=0

  while test $# -ge 1; do
    case $1 in
      -h|--help)
        echo "usage: $name [option]"
        echo ""
        echo "options:"
        echo "	-h|--help		display this message."
        echo "	-k i|--keep i		keep intermediate files."
        echo "	-v i|--verbose i	enable the verbose mode."
        exit 0;
        ;;
      -k|--keep)
        keep=$2;
        shift; shift;
        ;;
      -v|--verbose)
        verbose=$2;
        shift; shift;
        ;;
      *) shift;;
    esac;
  done

  if_verbose_2 echo "test: $name"
  test_fold "$name"
  if_verbose_2 "report status"
  report_status
}

main $@;
