package mail;

import java.util.*;


/*
 *  An enumeration type
 */

public class MailPriority implements Comparable {

  public static final MailPriority LOW = new MailPriority("Low");
  public static final MailPriority MEDIUM = new MailPriority("Medium");
  public static final MailPriority HIGH = new MailPriority("High");
  public static final MailPriority VERY_HIGH = new MailPriority("Very high");

  public String toString() {
    return name;
  }

  public int getOrdinal() {
    return ordinal;
  }

  public int compareTo(Object o) {
    if (o instanceof MailPriority) {
      return ordinal - ((MailPriority) o).getOrdinal();
    }
    return 0;
  }

  private MailPriority(String name) {
    this.name = name;
  }

  private static int nextOrdinal = 0;
  private final String name;
  private final int ordinal = nextOrdinal++;

}


public abstract class MailboxItem {

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public MailFolder getOwner() {
    return owner;
  }

  public void setOwner(MailFolder owner) {
    this.owner = owner;
  }

  public String toString() {
    return name;
  }

  public abstract int count();
  public abstract int countNewMail();

  protected MailboxItem(String name, MailFolder owner) {
    this.name = name;
    this.owner = owner;
  }

  protected String name;
  protected MailFolder owner;

}

/*
 *  An enumeration type
 */
public class MailStatus implements Comparable {

  public static final MailStatus NEW = new MailStatus("New");
  public static final MailStatus READ = new MailStatus("Read");
  public static final MailStatus REPLIED = new MailStatus("Replied");
  public static final MailStatus FORWARDED = new MailStatus("Forwarded");

  public String toString() {
    return name;
  }

  public int getOrdinal() {
    return ordinal;
  }

  public int compareTo(Object o) {
    if (o instanceof MailStatus) {
      return ordinal - ((MailStatus) o).getOrdinal();
    }
    return 0;
  }

  private MailStatus(String name) {
    this.name = name;
  }

  private static int nextOrdinal = 0;
  private final String name;
  private final int ordinal = nextOrdinal++;

}





public class MailFolder extends MailboxItem {

  Mail mail;

  public MailFolder(String name) {
    super(name, null);
  }

  public MailFolder(String name, MailFolder owner) {
    super(name, owner);
  }

  public List getItems() {
    return items;
  }

  public List getSubFolders() {
    List folders = new ArrayList();
    Iterator iterator = items.iterator();
    while (iterator.hasNext()) {
      Object item = iterator.next();
      if (item instanceof MailFolder) {
	folders.add(item);
      }
    }
    return folders;
  }

  public List getMails() {
    List mails = new ArrayList();
    Iterator iterator = items.iterator();
    while (iterator.hasNext()) {
      Object item = iterator.next();
      if (item instanceof Mail) {
	mails.add(item);
      }
    }
    return mails;
  }

  public void add(MailboxItem item) {
    items.add(item);
    item.setOwner(this);
  }

  public void add(MailboxItem[] items) {
    if (items != null) {
      for (int i = 0; i < items.length; i++) {
	add(items[i]);
      }
    }
  }

  public int count() {
    int count = 0;
    Iterator iterator = items.iterator();
    while (iterator.hasNext()) {
      Object item = iterator.next();
      if (item instanceof MailboxItem) {
	count += ((MailboxItem) item).count();
      }
    }
    return count;
  }

  public int countNewMail() {
    int count = 0;
    Iterator iterator = items.iterator();
    while (iterator.hasNext()) {
      Object item = iterator.next();
      if (item instanceof MailboxItem) {
	count += ((MailboxItem) item).countNewMail();
      }
    }
    return count;
  }

  protected List items = new ArrayList();

}




public class Mail extends MailboxItem {

  public Mail(String from,
	      String subject,
	      Date date,
	      MailPriority priority,
	      MailStatus status) {
    this(from, subject, null, date, priority, status, null, null);
  }

  public Mail(String from,
	      String subject,
	      MailFolder owner,
	      Date date,
	      MailPriority priority,
	      MailStatus status,
	      String message,
	      List attachments) {
    super(subject, owner);
    this.from = from;
    this.date = date;
    this.priority = priority;
    this.status = status;
    this.message = message;
    this.attachments = attachments;
  }

  public String getSubject() {
    return name;
  }

  public void setSubject(String subject) {
    name = subject;
  }

  public Date getDate() {
    return date;
  }

  public void setDate(Date date) {
    this.date = date;
  }

  public MailPriority getPriority() {
    return priority;
  }

  public void setPriority(MailPriority priority) {
    this.priority = priority;
  }

  public MailStatus getStatus() {
    return status;
  }

  public void setStatus(MailStatus status) {
    this.status = status;
  }

  public String getFrom() {
    return from;
  }

  public void setFrom(String from) {
    this.from = from;
  }

  public String getMessage() {
    return message;
  }

  public void setMessage(String message) {
    this.message = message;
  }

  public List getAttachments() {
    return attachments;
  }

  public void addAttachment(Object attachment) {
    if (attachment != null) {
      attachments.add(attachment);
    }
  }

  public int count() {
    return 1;
  }

  public int countNewMail() {
    return (status == MailStatus.NEW ? 1 : 0);
  }

  public String toString() {
    return "From: " + from + "; Subject: " + name + ";\n  Received: " + date +
      "; Priority: " + priority + "; Status: " + status + ";";
  }

  protected Date date;
  protected MailPriority priority;
  protected MailStatus status;
  protected String from;
  protected String message;
  protected List attachments;

}

public class Main {

    /*
     * The following code is due to some Dryad type checker bugs.
     */

  static MailPriority very_high = MailPriority.VERY_HIGH;
  static MailPriority high = MailPriority.HIGH;
  static MailPriority medium = MailPriority.MEDIUM;
  static MailPriority low = MailPriority.LOW;

  static MailStatus new_ = MailStatus.NEW;
  static MailStatus read = MailStatus.READ;
  static MailStatus replied = MailStatus.REPLIED;
  static MailStatus forwarded = MailStatus.FORWARDED;

  static Date[] times = {
	     getTime(2001, Calendar.NOVEMBER, 20, 23, 50),
	     getTime(2001, Calendar.NOVEMBER, 19, 23, 39),
	     getTime(2001, Calendar.NOVEMBER, 18, 14, 56),
	     getTime(2001, Calendar.OCTOBER, 9, 23, 30),
	     getTime(2001, Calendar.OCTOBER, 2, 22, 40),
	     getTime(2001, Calendar.NOVEMBER, 16, 20, 10),
	     getTime(2001, Calendar.OCTOBER, 24, 4, 21),
	     getTime(2001, Calendar.DECEMBER, 2, 9, 24),
	     getTime(2002, Calendar.JULY, 6, 7, 00),
	     getTime(2003, Calendar.JANUARY, 1, 0, 00),
	     getTime(2001, Calendar.SEPTEMBER, 21, 9, 30),
	     getTime(2001, Calendar.SEPTEMBER, 11, 6, 23),
	     getTime(2001, Calendar.OCTOBER, 26, 22, 12),
	     getTime(2001, Calendar.OCTOBER, 2, 23, 59)};

  static Mail[] se450 = {
    new Mail("Bill Gates", "Need extension for final project",
	     times[0],
	     very_high, replied),
    new Mail("Linus Torvalds", "Final project: Linux -- a small OS",
             times[1],
	     very_high, replied),
    new Mail("John Valissides", "Question on Facade pattern",
	     times[2],
	     very_high, replied),
  };

  static Mail[] se452 = {
    new Mail("Marc Andreesen", "Problem with Netscape 7.02",
	     times[3],
	     high, new_),
    new Mail("James Gosling", "HW4 submission, version 3.14159",
	     times[4],
	     high, new_),
  };

  static Mail[] work = {
    new Mail("Ralph Johnson", "CFP: PLOP",
	     times[5],
	     high, new_),
    new Mail("Chris Galvin", "Your proposal",
	     times[6],
	     high, new_),
    new Mail("David Boise", "Abstract of my talk",
	     times[7],
	     high, new_),
  };

  static Mail[] news = {
    new Mail("WSJ.com", "Yahoo files for bankruptcy",
	     times[8],
	     high, new_),
    new Mail("CERT", "Security alert, Windows XXP-3",
	     times[9],
	     very_high, new_),
    new Mail("Scott McNealy", "Top 10 reasons for breaking up Microsoft",
	     times[10],
	     medium, read),
    new Mail("Gordon Moore", "10 myths of Moors's law",
	     times[11],
	     medium, read),
  };

  static Mail[] junks = {
    new Mail("William Gates, III", ".NET Framework",
	     times[11],
	     high, new_),
    new Mail("AOL", "Free unlimited DSL with ...",
	     times[12],
	     high, new_),
  };

  public static MailFolder buildInbox() {
    MailFolder inboxFolder = new MailFolder("Inbox");
    MailFolder coursesFolder = new MailFolder("Courses");
    MailFolder se450Folder = new MailFolder("SE450");
    se450Folder.add(se450);
    MailFolder se452Folder = new MailFolder("SE452");
    se452Folder.add(se452);
    coursesFolder.add(se450Folder);
    coursesFolder.add(se452Folder);
    MailFolder workFolder = new MailFolder("Work");
    workFolder.add(work);
    MailFolder newsFolder = new MailFolder("News");
    newsFolder.add(news);
    MailFolder junksFolder = new MailFolder("Junk Mails");
    junksFolder.add(junks);
    inboxFolder.add(coursesFolder);
    inboxFolder.add(workFolder);
    inboxFolder.add(newsFolder);
    inboxFolder.add(junksFolder);
    return inboxFolder;
  }

  public static void main(String[] args) {
    MailFolder inboxFolder = buildInbox();
    System.out.println("You " + inboxFolder.count() + " mails in Inbox");
    System.out.println("You " + inboxFolder.countNewMail() + " new mails in Inbox");
  }

  protected static Date getTime(int year, int month, int date, int hour, int minute) {
    Calendar calendar = Calendar.getInstance();
    calendar.set(year, month, date, hour, minute, 0);
    return calendar.getTime();
  }

}

