package maze;

import java.awt.*;
import java.applet.AudioClip;
import java.util.Stack;
import java.util.List;
import java.util.ArrayList;
import java.awt.event.*;
import javax.swing.*;
import java.applet.*;
import java.net.*;



public interface Command {

  public void execute();

}

/*
 *  An enumeration type
 */
public class Orientation {

  public static final Orientation VERTICAL = new Orientation("Vertical");
  public static final Orientation HORIZONTAL = new Orientation("Horizontal");

  public String toString() {
    return name;
  }

  private Orientation(String name) {
    this.name = name;
  }

  private final String name;

}




/*
 *  An enumeration type
 */


public class Direction //implements Comparable
{

  public static final Direction NORTH = new Direction("North");
  public static final Direction EAST  = new Direction("East");
  public static final Direction SOUTH = new Direction("South");
  public static final Direction WEST  = new Direction("West");

  public String toString() {
    return name;
  }

  public int getOrdinal() {
    return ordinal;
  }

  public int compareTo(Object o) {
    if (o instanceof Direction) {
      return ordinal - ((Direction) o).getOrdinal();
    }
    return 0;
  }

  public static Direction first() {
    return values[0];
  }

  public Direction next() {
    if (ordinal < values.length - 1) {
      return values[ordinal + 1];
    } else {
      return null;
    }
  }

  public Direction opposite() {
    return values[(ordinal + 2) % 4];
  }

  /*
  public static Direction first() {
    return NORTH;
  }

  public Direction next() {
    if (this == NORTH) {
      return EAST;
    } else if (this == EAST) {
      return SOUTH;
    } else if (this == SOUTH) {
      return WEST;
    } else if (this == WEST) {
      return null;
    }
    return null;
  }

  public Direction opposite() {
    if (this == NORTH) {
      return SOUTH;
    } else if (this == EAST) {
      return WEST;
    } else if (this == SOUTH) {
      return NORTH;
    } else if (this == WEST) {
      return EAST;
    }
    return null;
  }
  */

  private Direction(String name) {
    this.name = name;
  }

  private static int nextOrdinal = 0;
  private final String name;
  private final int ordinal = nextOrdinal++;

/* Salima:Probleme de déclarations de tableau:ne compile pas */
private static final Direction values ;
// private static final Direction[] values = {new Direction.NORTH,new Direction.EAST, new Direction.SOUTH, new Direction.WEST };

}



public class AudioUtility {

  public static final AudioClip getAudioClip(String filename) {
    if (filename != null) {
      //try {
	return Applet.newAudioClip(new URL("file:" + filename));
      //} catch (MalformedURLException e) {
	//System.err.println(e.getMessage());
      //}
    }
    return null;
  }

}

public interface MapSite extends Cloneable {

  public Object clone() throws CloneNotSupportedException;
  public void enter(Maze maze);
  public void draw(Graphics g, int x, int y, int w, int h);

}



public class Door implements MapSite {

  public Door(Room room1, Room room2) {
    this.room1 = room1;
    this.room2 = room2;
  }

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  public boolean isOpen() {
    return open;
  }

  public void setOpen(boolean open) {
    this.open = open;
  }

  public void setRooms(Room room1, Room room2) {
    this.room1 = room1;
    this.room2 = room2;
  }

  public Orientation getOrientation() {
    return orientation;
  }

  public void setOrientation(Orientation orientation) {
    this.orientation = orientation;
  }

  public Room otherSideFrom(Room room) {
    if (room != null) {
      if (room == room1) {
	return room2;
      } else if (room == room2) {
	return room1;
      }
    }
    return null;
  }

  public void enter(Maze maze) {
    if (open) {
      Room otherRoom = otherSideFrom(maze.getCurrentRoom());
      if (otherRoom != null) {
	otherRoom.enter(maze);
      }
    } else {
      ding.play();
    }
  }

  public void draw(Graphics g, int x, int y, int w, int h) {
    g.setColor(Wall.WALL_COLOR);
    g.fillRect(x, y, w, h);
    if (orientation == Orientation.VERTICAL) {
      y += 2 * w; h -= 4 * w;
    } else {
      x += 2 * h; w -= 4 * h;
    }
    if (open) {
      g.setColor(Room.ROOM_COLOR);
      g.fillRect(x, y, w, h);
    } else {
      g.setColor(Color.red);
      g.fillRect(x, y, w, h);
      g.setColor(Color.black);
      g.drawRect(x, y, w, h);
    }
  }

  protected Room room1;
  protected Room room2;
  protected boolean open;
  protected Orientation orientation;

  protected static AudioClip ding = AudioUtility.getAudioClip("audio/ding.au");

}



public class Room implements MapSite {

  public static final Color ROOM_COLOR = new Color(152, 251, 152);
  public static final Color PLAYER_COLOR = Color.red;

  public Room(int roomNumber) {
    this.roomNumber = roomNumber;
  }

  public Object clone() throws CloneNotSupportedException {
    Room room = (Room) super.clone();
    room.sides = new MapSite[4];
    for (int i = 0; i < 4; i++) {
      if (sides[i] != null) {
	room.sides[i] = (MapSite) sides[i].clone();
      }
    }
    return room;
  }

  public MapSite getSide(Direction dir) {
    if (dir != null) {
      return sides[dir.getOrdinal()];
    }
    return null;
  }

  public void setSide(Direction dir, MapSite site) {
    if (dir != null) {
      sides[dir.getOrdinal()] = site;
      if (site instanceof Door) {
	Door door = (Door) site;
	if (dir == Direction.NORTH ||
	    dir == Direction.SOUTH) {
	  door.setOrientation(Orientation.HORIZONTAL);
	} else {
	  door.setOrientation(Orientation.VERTICAL);
	}
      }
    }
  }

  public void setRoomNumber(int roomNumber) {
    this.roomNumber = roomNumber;
  }

  public int getRoomNumber() {
    return roomNumber;
  }

  public Point getLocation() {
    return location;
  }

  public void setLocation(Point location) {
    this.location = location;
  }

  public void enter(Maze maze) {
    maze.setCurrentRoom(this);
    gong.play();
  }

  public boolean isInRoom() {
    return inroom;
  }

  public void setInRoom(boolean inroom) {
    this.inroom = inroom;
  }

  public void draw(Graphics g, int x, int y, int w, int h) {
    g.setColor(ROOM_COLOR);
    g.fillRect(x, y, w, h);
    if (inroom) {
      g.setColor(PLAYER_COLOR);
      g.fillOval(x + w / 2 - 5, y + h / 2 - 5, 10, 10);
    }
  }

  protected int roomNumber = 0;
  protected boolean inroom = false;
  protected MapSite[] sides = new MapSite[4];
  protected Point location = null;

  protected static AudioClip gong = AudioUtility.getAudioClip("audio/gong.au");
  //protected static AudioClip spacemusic = util.AudioUtility.getAudioClip("audio/spacemusic.au");
}


public class Wall implements MapSite {

  public static final Color WALL_COLOR = Color.orange;

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  public void enter(Maze maze) {
    hurts.play();
  }

  public void draw(Graphics g, int x, int y, int w, int h) {
    g.setColor(WALL_COLOR);
    g.fillRect(x, y, w, h);
  }

  protected static AudioClip hurts = AudioUtility.getAudioClip("audio/that.hurts.au");

}




public class Maze implements Cloneable {

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  public void addRoom(Room room) {
    if (room != null) {
      rooms.add(room);
    }
  }

  public Room findRoom(int roomNumber) {
    for (int i = 0; i < rooms.size(); i++) {
      Room room = (Room) rooms.get(i);
      if (roomNumber == room.getRoomNumber()) {
	return room;
      }
    }
    return null;
  }

  public void setCurrentRoom(int roomNumber) {
    Room room = findRoom(roomNumber);
    setCurrentRoom(room);
  }

  public void setCurrentRoom(Room room) {
    if (room != curRoom) {
      if (curRoom != null) {
	curRoom.setInRoom(false);
      }
      if (room != null) {
	room.setInRoom(true);
	curRoom = room;
      }
      if (view != null) {
	view.repaint();
      }
    }
  }

  public Room getCurrentRoom() {
    return curRoom;
  }

  public void move(Direction direction) {
    if (curRoom != null) {
      MapSite side = curRoom.getSide(direction);
      if (side != null) {
	side.enter(this);
      }
    }
  }

  public void draw(Graphics g) {
    if (dim == null) {
      calculateDimension();
    }
    int dx = MARGIN + -offset.x * ROOM_SIZE;
    int dy = MARGIN + -offset.y * ROOM_SIZE;

    if (debug) {
      System.out.println("Maze.Draw(): offset=" + offset.x + ", " + offset.y);
    }

    // draw rooms first
    for (int i = 0; i < rooms.size(); i++) {
      Room room = (Room) rooms.get(i);
      if (room != null) {
	Point location = room.getLocation();
	if (location != null) {

	  if (debug) {
	    System.out.println("Maze.Draw(): Room " + room.getRoomNumber() +
			       " location: " + location.x + ", " + location.y);
	  }

	  room.draw(g,
		    dx + location.x * ROOM_SIZE,
		    dy + location.y * ROOM_SIZE,
		    ROOM_SIZE, ROOM_SIZE);
	}
      }
    }
    // draw walls and doors
    for (int i = 0; i < rooms.size(); i++) {
      Room room = (Room) rooms.get(i);
      if (room != null) {
	Point location = room.getLocation();
	if (location != null) {
	  for (Direction dir = Direction.first(); dir != null; dir = dir.next()) {
	    MapSite side = room.getSide(dir);
	    if (side != null) {
	      if (dir == Direction.NORTH) {
		side.draw(g,
			  dx + location.x * ROOM_SIZE - WALL_THICKNESS / 2,
			  dy + location.y * ROOM_SIZE - WALL_THICKNESS / 2,
			  ROOM_SIZE + WALL_THICKNESS,
			  WALL_THICKNESS);
	      } else if (dir == Direction.EAST) {
		side.draw(g,
			  dx + location.x * ROOM_SIZE + ROOM_SIZE - WALL_THICKNESS / 2,
			  dy + location.y * ROOM_SIZE - WALL_THICKNESS / 2,
			  WALL_THICKNESS,
			  ROOM_SIZE + WALL_THICKNESS);
	      } else if (dir == Direction.SOUTH) {
		side.draw(g,
			  dx + location.x * ROOM_SIZE - WALL_THICKNESS / 2,
			  dy + location.y * ROOM_SIZE + ROOM_SIZE - WALL_THICKNESS / 2,
			  ROOM_SIZE + WALL_THICKNESS,
			  WALL_THICKNESS);
	      } else {
		side.draw(g,
			  dx + location.x * ROOM_SIZE - WALL_THICKNESS / 2,
			  dy + location.y * ROOM_SIZE - WALL_THICKNESS / 2,
			  WALL_THICKNESS,
			  ROOM_SIZE + WALL_THICKNESS);
	      }
	    }
	  }
	}
      }
    }
  }

  public Dimension getDimension() {
    if (dim == null) {
      calculateDimension();
    }
    return dim;
  }

  protected void calculateDimension() {
    if (rooms.size() > 0) {
      int minX = 0, maxX = 0, minY = 0, maxY = 0;
      Room room = (Room) rooms.get(0);
      room.setLocation(new Point(0, 0));
      boolean changed = true;
      while (changed &&
	     !isAllRoomsSet()) {
	changed = false;
	for (int i = 0; i < rooms.size(); i++) {
	  room = (Room) rooms.get(i);
	  Point location = room.getLocation();
	  if (location != null) {
	    for (Direction dir = Direction.first(); dir != null; dir = dir.next()) {
	      MapSite side = room.getSide(dir);
	      if (side instanceof Door) {
		Door door = (Door) side;
		Room otherSide = door.otherSideFrom(room);
		if (otherSide != null &&
		    otherSide.getLocation() == null) {
		  if (dir == Direction.NORTH) {
		    otherSide.setLocation(new Point(location.x, location.y - 1));
		    minY = Math.min(minY, location.y - 1);
		  } else if (dir == Direction.EAST) {
		    otherSide.setLocation(new Point(location.x + 1, location.y));
		    maxX = Math.max(maxX, location.x + 1);
		  } else if (dir == Direction.SOUTH) {
		    otherSide.setLocation(new Point(location.x, location.y + 1));
		    maxY = Math.max(maxY, location.y + 1);
		  } else {
		    otherSide.setLocation(new Point(location.x - 1, location.y));
		    minX = Math.min(minX, location.x - 1);
		  }
		  changed = true;
		}
	      }
	    }
	  }
	}
      }
      offset = new Point(minX, minY);
      dim = new Dimension(maxX - minX + 1, maxY - minY + 1);
    } else {
      offset = new Point(0, 0);
      dim = new Dimension(0, 0);
    }
  }

  protected boolean isAllRoomsSet() {
    for (int i = 0; i < rooms.size(); i++) {
      Room room = (Room) rooms.get(i);
      if (room.getLocation() == null) {
	return false;
      }
    }
    return true;
  }

  protected void setView(Component view) {
    this.view = view;
  }

  protected void doCommand(Command command) {
    if (command != null) {
      moves.push(command);
      command.execute();
    }
  }

  protected void undoCommand() {
    if (!moves.empty()) {
      Object top = moves.peek(); // looking at the top element without popping it
      if (top instanceof UndoableCommand) {
	moves.pop();
	UndoableCommand undoableCommand = (UndoableCommand) top;
	undoableCommand.undo();
      }
    }
  }

  protected List rooms = new ArrayList();
  protected Dimension dim;
  protected Point offset;
  protected Room curRoom = null;
  protected Stack moves = new Stack();

  protected Component view;

  private static final int ROOM_SIZE = 40;
  private static final int WALL_THICKNESS = 6;
  private static final int MARGIN = 20;

  private static final boolean debug = true;

  protected void showFrame(String frameTitle) {
    JFrame frame;
    frame = new JFrame(frameTitle);
//    frame.setContentPane(new Maze.MazePanel(this));
    frame.pack();
    Dimension frameDim = frame.getSize();
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    frame.setLocation(screenSize.width / 2 - frameDim.width / 2,
		      screenSize.height / 2 - frameDim.height / 2);
    frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
    frame.setVisible(true);
  }

  public static class MazePanel extends JPanel {

    public MazePanel(Maze maze) {
      this.maze = maze;
      if (maze != null) {
	maze.setView(this);
	Dimension d = maze.getDimension();
	if (d != null) {
	  dim = new Dimension(d.width * ROOM_SIZE + 2 * MARGIN,
			      d.height * ROOM_SIZE + 2 * MARGIN);
	}

//	addKeyListener(new MazeKeyListener(maze));

      }
    }

    public void paint(Graphics g) {
      Dimension d = getSize();
      g.setColor(Color.white);
      g.fillRect(0, 0, d.width, d.height);
      if (maze != null) {
	maze.draw(g);
      }
      requestFocus();
    }

    //public boolean isFocusTraversable() { // pre 1.4
    public boolean isFocusable() { // 1.4
      return true;
    }

    public Dimension getPreferredSize() {
      return dim;
    }

    public Dimension getMinimumSize() {
      return dim;
    }

    private Maze maze;
    private Dimension dim;

  }

  static class MazeKeyListener extends KeyAdapter {

    MazeKeyListener(Maze maze) {
      this.maze = maze;
    }

    public void keyPressed(KeyEvent e) {
      System.out.println("Key pressed");
      Command command = null;
      int code = e.getKeyCode();
      switch (code) {
      case KeyEvent.VK_UP:
	System.out.println("Up key");
	command = new MazeMoveCommand(maze, Direction.NORTH);
	break;
      case KeyEvent.VK_DOWN:
	System.out.println("Down key");
	command = new MazeMoveCommand(maze, Direction.SOUTH);
	maze.move(Direction.SOUTH);
	break;
      case KeyEvent.VK_LEFT:
	System.out.println("Left key");
	command = new MazeMoveCommand(maze, Direction.WEST);
	break;
      case KeyEvent.VK_RIGHT:
	System.out.println("Right key");
	command = new MazeMoveCommand(maze, Direction.EAST);
	break;
      default:
	System.out.println("Key press ignored");
      }
      if (command != null) {
	maze.doCommand(command);
      }
    }

    Maze maze;
  }

}


public interface UndoableCommand extends Command {

  public void undo();

}



public class MazeMoveCommand implements UndoableCommand {

  public MazeMoveCommand(Maze maze, Direction direction) {
    this.maze = maze;
    this.direction = direction;
  }

  public void execute() {
    maze.move(direction);
  }

  public void undo() {
    maze.move(direction.opposite());
  }

  protected Maze maze;
  protected Direction direction;

}




/*
 *  Build a Maze game.
 *
 *  This implmentation does not use any design patterns.
 *  Compare the createMaze() method of this class to the ccorresponding method of the following
 *  classes which create the same Maze using different creational design patterns.
 *   - MazeGameAbstractFactory: using Abstract Factory design pattern
 *   - MazeGameBuilder: using Builder design pattern
 *
 *  Run the program as follows:
 *   java maze.SimpleMazeGame Large
 *         -- create a large Maze game
 *   java maze.SimpleMazeGame
 *         -- create a small Maze game
 */
public class SimpleMazeGame {

  /**
   *  Creates a small maze with 2 rooms.
   */
  public static Maze createMaze() {
    Maze maze = new Maze();
    Room room1 = new Room(1);
    Room room2 = new Room(2);
    Door door = new Door(room1, room2);

    room1.setSide(Direction.NORTH, new Wall());
    room1.setSide(Direction.EAST, door);
    room1.setSide(Direction.SOUTH, new Wall());
    room1.setSide(Direction.WEST, new Wall());

    room2.setSide(Direction.NORTH, new Wall());
    room2.setSide(Direction.EAST, new Wall());
    room2.setSide(Direction.SOUTH, new Wall());
    room2.setSide(Direction.WEST, door);

    maze.addRoom(room1);
    maze.addRoom(room2);

    return maze;
  }

  /**
   *  Creates a large maze with 9 rooms.
   */
  public static Maze createLargeMaze() {
    Maze maze = new Maze();
    Room room1 = new Room(1);
    Room room2 = new Room(2);
    Room room3 = new Room(3);
    Room room4 = new Room(4);
    Room room5 = new Room(5);
    Room room6 = new Room(6);
    Room room7 = new Room(7);
    Room room8 = new Room(8);
    Room room9 = new Room(9);
    Door door1 = new Door(room1, room2);
    Door door2 = new Door(room2, room3);
    Door door3 = new Door(room4, room5);
    Door door4 = new Door(room5, room6);
    Door door5 = new Door(room5, room8);
    Door door6 = new Door(room6, room9);
    Door door7 = new Door(room7, room8);
    Door door8 = new Door(room1, room4);

    door1.setOpen(true);
    door2.setOpen(false);
    door3.setOpen(true);
    door4.setOpen(true);
    door5.setOpen(false);
    door6.setOpen(true);
    door7.setOpen(true);
    door8.setOpen(true);

    room1.setSide(Direction.NORTH, door8);
    room1.setSide(Direction.EAST, new Wall());
    room1.setSide(Direction.SOUTH, new Wall());
    room1.setSide(Direction.WEST, door1);

    room2.setSide(Direction.NORTH, new Wall());
    room2.setSide(Direction.EAST, door1);
    room2.setSide(Direction.SOUTH, new Wall());
    room2.setSide(Direction.WEST, door2);

    room3.setSide(Direction.NORTH, new Wall());
    room3.setSide(Direction.EAST, door2);
    room3.setSide(Direction.SOUTH, new Wall());
    room3.setSide(Direction.WEST, new Wall());

    room4.setSide(Direction.NORTH, new Wall());
    room4.setSide(Direction.EAST, new Wall());
    room4.setSide(Direction.SOUTH, door8);
    room4.setSide(Direction.WEST, door3);

    room5.setSide(Direction.NORTH, door5);
    room5.setSide(Direction.EAST, door3);
    room5.setSide(Direction.SOUTH, new Wall());
    room5.setSide(Direction.WEST, door4);

    room6.setSide(Direction.NORTH, door6);
    room6.setSide(Direction.EAST, door4);
    room6.setSide(Direction.SOUTH, new Wall());
    room6.setSide(Direction.WEST, new Wall());

    room7.setSide(Direction.NORTH, new Wall());
    room7.setSide(Direction.EAST, new Wall());
    room7.setSide(Direction.SOUTH, new Wall());
    room7.setSide(Direction.WEST, door7);

    room8.setSide(Direction.NORTH, new Wall());
    room8.setSide(Direction.EAST, door7);
    room8.setSide(Direction.SOUTH, door5);
    room8.setSide(Direction.WEST, new Wall());

    room9.setSide(Direction.NORTH, new Wall());
    room9.setSide(Direction.EAST, new Wall());
    room9.setSide(Direction.SOUTH, door6);
    room9.setSide(Direction.WEST, new Wall());

    maze.addRoom(room1);
    maze.addRoom(room2);
    maze.addRoom(room3);
    maze.addRoom(room4);
    maze.addRoom(room5);
    maze.addRoom(room6);
    maze.addRoom(room7);
    maze.addRoom(room8);
    maze.addRoom(room9);

    return maze;
  }

  public static void main(String[] args) {
    Maze maze;
    if (args.length > 0 &&
	"Large".equals(args[0])) {
      maze = createLargeMaze();
    } else {
      maze = createMaze();
    }
    maze.setCurrentRoom(1);
    maze.showFrame("Maze");
  }


}
